# This create a canvas with a xy-diagram (x-axis can be linear or logarithmical)
#
# usage
#
# ::FR_diagram2::create
# --------------------
#
# create a 2D diagram; can also draw datas and show a diagram title in one step
# -path            {path / frame for the diagram}
# -x_axis_line_cnt {count of vertical grid lines | default: 5}
# -y_axis_line_cnt {count of horizontal grid lines | default: 11}
# -x_axis_min      {minimum value of the x-axis | default: 0}
# -x_axis_max      {maximum value of the x-axis | default: 1}
# -y_axis_min      {minimum value of the y-axis | default: -1}
# -y_axis_max      {maximum value of the y-axis | default: 1}
# -data            {list with xy-values in the from x1 y1 x2 y2...}
# -x_axis_log      {set 1 if you want to have a logarithmical x-axis | default: 0}
# -diagram_title   {diagram title | default: ""}
#
# returns an id of the created diagram (needed for further handling)
#
# ::FR_diagram2::draw_data
# -----------------------
# 
# erase the old curve and draw new datas (linear x-axis)
# ::FR_diagram2::draw_data {DiagramId Data}
#
# ::FR_diagram2::draw_log_data
# -----------------------
# 
# erase the old curve and draw new datas (logarithmical x-axis)
# ::FR_diagram2::draw_log_data {DiagramId Data}
#
#
# ::FR_diagram2::diagram_title
# ---------------------------
# erase the old and set a new diagram title
# ::FR_diagram2::diagram_title {DiagramId String}
#
# ::FR_diagram2::zoom_out
# ---------------------------
# go back to the last zoom (view_all=0) or 
# draw the whole curve (view_all=1)
# ::FR_diagram2::zoom_out {DiagramId view_all}
#

package provide FR_diagram2 0.1

namespace eval ::FR_diagram2 \
{
 namespace export create diagram_title draw_data zoom_out
}


# ---------------------------------------------- 
# the start point
# ---------------------------------------------- 
proc ::FR_diagram2::create {args} \
{ 
 variable DATA; variable VAR; variable ID
 
 # generate a new diagram id
 if {[info exists ID] == 0} \
 {
   set ID 0 
 } else \
 {
   incr ID
 }
 
 # we set the default values
 set VAR($ID,Path) ""
 set VAR($ID,x_axis_line_cnt) 5
 set VAR($ID,y_axis_line_cnt) 11
 set VAR($ID,x_axis_log) 0
 set VAR($ID,x_data_min) 0
 set VAR($ID,x_data_max) 1
 set VAR($ID,y_data_min) -1
 set VAR($ID,y_data_max) 1
 set VAR($ID,diagram_title) ""
 set DATA($ID) ""
 set VAR($ID,data_range) ""
 set VAR($ID,redraw_request) ""
 
 for {set i 0} {$i < [llength $args]} {incr i 2} \
 {
  set arg [lindex $args $i]

  switch -glob -- $arg \
  {-path            {set VAR($ID,Path) [lindex $args [expr {$i+1}]]} \
   -x_axis_line_cnt {set VAR($ID,x_axis_line_cnt) [lindex $args [expr {$i+1}]]} \
   -y_axis_line_cnt {set VAR($ID,y_axis_line_cnt) [lindex $args [expr {$i+1}]]} \
   -x_axis_min      {set VAR($ID,x_data_min) [lindex $args [expr {$i+1}]]} \
   -x_axis_max      {set VAR($ID,x_data_max) [lindex $args [expr {$i+1}]]} \
   -y_axis_min      {set VAR($ID,y_data_min) [lindex $args [expr {$i+1}]]} \
   -y_axis_max      {set VAR($ID,y_data_max) [lindex $args [expr {$i+1}]]} \
   -x_axis_log      {set VAR($ID,x_axis_log) [lindex $args [expr {$i+1}]]} \
   -diagram_title   {set VAR($ID,diagram_title) [lindex $args [expr {$i+1}]]} \
   -data            {set DATA($ID) [lindex $args [expr {$i+1}]]} \
   --               {break} \
   *                {puts "Unknown option: $arg"; exit 1}
   } 
  }
 
 # we need a path for the canvas widget
 if {$VAR($ID,Path) eq ""} {error "You must set a path name for your diagram with \"-path\""}
 set VAR($ID,Path) [string trimleft $VAR($ID,Path) .]
 
 # we need at least 2 lines: for the minimum and the maximum value
 if {$VAR($ID,x_axis_line_cnt) < 2} {set VAR($ID,x_axis_line_cnt) 2}
 if {$VAR($ID,y_axis_line_cnt) < 2} {set VAR($ID,y_axis_line_cnt) 2}
 
 ::FR_diagram2::init $ID
 ::FR_diagram2::look $ID
 
 ::FR_diagram2::draw_y_axis $ID $VAR($ID,y_data_min) $VAR($ID,y_data_max)
 
 if {$VAR($ID,x_axis_log)} \
 {
  ::FR_diagram2::draw_log_x_axis $ID $VAR($ID,x_data_min) $VAR($ID,x_data_max)
 } else \
 {
  ::FR_diagram2::draw_x_axis $ID $VAR($ID,x_data_min) $VAR($ID,x_data_max)
 }
 
 # we draw the curve
 set data_length [llength $DATA($ID)]

 if {$data_length > 3} \
 { 
  set VAR($ID,last_element) [expr {$data_length - 1}]
  ::FR_diagram2::draw_curve $ID [list 0 $VAR($ID,last_element)]
  
  # we preset the zoom history with the whole plot
  set VAR($ID,zoom_history) [list 0 $VAR($ID,last_element)] 
 }
 
 if {$VAR($ID,diagram_title) ne ""} \
 {
  ::FR_diagram2::diagram_title $ID
 }

 return $ID
}


# ---------------------------------------------- 
# we create a canvas and set the bindings
# ---------------------------------------------- 
proc ::FR_diagram2::init {Id} \
{
 variable VAR

 set VAR($Id,Can) "$VAR($Id,Path).can"

 if {[winfo exists .$VAR($Id,Can)]} {destroy .$VAR($Id,Can)}
 
 canvas .$VAR($Id,Can) -background white
 pack   .$VAR($Id,Can) -fill both -expand yes
 
 # the menu for the zoom operations
 menu .$VAR($Id,Can).menu -tearoff 0
 .$VAR($Id,Can).menu add command -label [msgcat::mc "Zoom out"] -command [list ::FR_diagram2::zoom_out $Id 0]
 .$VAR($Id,Can).menu add separator
 .$VAR($Id,Can).menu add command -label [msgcat::mc "View all"] -command [list ::FR_diagram2::zoom_out $Id 1]
 
 bind .$VAR($Id,Can) <ButtonPress-1> [list ::FR_diagram2::start_zoom_box $Id %x %y]
 bind .$VAR($Id,Can) <Motion> [list ::FR_diagram2::draw_zoom_box $Id %x %y]
 bind .$VAR($Id,Can) <ButtonRelease-1> [list ::FR_diagram2::end_zoom_box $Id %x %y]
 bind .$VAR($Id,Can) <Button-3> [list tk_popup .$VAR($Id,Can).menu %X %Y]
 
 ::FR_diagram2::getsize $Id
 
 # we bind the parent of canvas to its Configure command
 # to redraw the canvas
 set canvas_parent [winfo parent .$VAR($Id,Can)]
 if {$canvas_parent eq ""} {set canvas_parent .}
 
 bind $canvas_parent <Configure> [list ::FR_diagram2::redraw_request $Id]
}


# ------------------------------------------------------------------------------
# we ask for the whole size of the canvas (we need this for the redraw function)
# ------------------------------------------------------------------------------ 
proc ::FR_diagram2::get_canvas_size {Id} \
{
 variable VAR
 
 set VAR($Id,canvas_x_size) [winfo width .$VAR($Id,Can)]
 set VAR($Id,canvas_y_size) [winfo height .$VAR($Id,Can)]
}


# ---------------------------------------------- 
# returns the internal x-size of the canvas
# ---------------------------------------------- 
proc ::FR_diagram2::getxsize {Id} \
{
 variable VAR
 
 set VAR($Id,cmx) [ expr {[winfo width .$VAR($Id,Can)]-2*[.$VAR($Id,Can) cget -border]-2*[.$VAR($Id,Can) cget -highlightthickness]} ]
 
 if {$VAR($Id,cmx) <= 1} {set VAR($Id,cmx) [.$VAR($Id,Can) cget -width]}
}


# ---------------------------------------------- 
# returns the internal y-size of the canvas
# ---------------------------------------------- 
proc ::FR_diagram2::getysize {Id} \
{
 variable VAR
 
 set VAR($Id,cmy) [ expr {[winfo height .$VAR($Id,Can)]-2*[.$VAR($Id,Can) cget -border]-2*[.$VAR($Id,Can) cget -highlightthickness]} ]
 
 if {$VAR($Id,cmy) <= 1} {set VAR($Id,cmy) [.$VAR($Id,Can) cget -height]}
}


# ---------------------------------------------------------------------
# we ask for the size of the canvas (inside, the useable painting area)
# ---------------------------------------------------------------------
proc ::FR_diagram2::getsize {Id} \
{
 update
 
 ::FR_diagram2::get_canvas_size $Id
 
 ::FR_diagram2::getxsize $Id
 ::FR_diagram2::getysize $Id
}


# ---------------------------------------------- 
# precalculation of the diagram size
# ---------------------------------------------- 
proc ::FR_diagram2::look {Id} \
{
 variable VAR

 # the spaces around the diagram
 set bottom 0.1
 set left   0.1
 set right  0.05
 set top    0.1
 
 set VAR($Id,y_axis_range) [expr {$VAR($Id,cmy) * (1 - $bottom - $top)}]
 set VAR($Id,delta_y_axis_lines) [expr {$VAR($Id,y_axis_range) / ($VAR($Id,y_axis_line_cnt) - 1)}]
 
 set VAR($Id,y_axis_text_xpos) [expr {$VAR($Id,cmx) * $left * 0.75}]
 set VAR($Id,y_axis_line_xpos) [expr {$VAR($Id,cmx) * (0.01 + $left)}]
 
 set VAR($Id,x_axis_range) [expr {$VAR($Id,cmx) * (1 - $left - $right)}]
 set VAR($Id,delta_x_axis_lines) [expr {$VAR($Id,x_axis_range) / ($VAR($Id,x_axis_line_cnt) - 1)}]
 
 set VAR($Id,x_axis_text_ypos) [expr {$VAR($Id,cmy) * (1 - $bottom * 0.5)}]
 set VAR($Id,x_axis_line_ypos) [expr {$VAR($Id,cmy) * (1 - (0.01 + $bottom))}]
 
 set VAR($Id,y_axis_min) [expr {$VAR($Id,cmy) * $top}]
 set VAR($Id,y_axis_max) [expr {$VAR($Id,cmy) * (1 - $bottom)}]
 
 set VAR($Id,x_axis_min) [expr {$VAR($Id,cmx) * $left}]
 set VAR($Id,x_axis_max) [expr {$VAR($Id,cmx) * (1 - $right)}]
 
 set VAR($Id,diagram_title_ypos) [expr {$VAR($Id,cmy) * $top * 0.5}]
 set VAR($Id,diagram_title_xpos) [expr {($VAR($Id,x_axis_range) / 2) + $VAR($Id,x_axis_min)}]
}


# -----------------------------------------------
# draws the linear y-axis and the horizontal grid
# -----------------------------------------------
proc ::FR_diagram2::draw_y_axis {Id y_min y_max {lettering_only 0}} \
{
 variable VAR
 
 set VAR($Id,y_data_min) $y_min
 set VAR($Id,y_data_max) $y_max
 set VAR($Id,y_data_range) [expr {$y_max - $y_min}]
 
 
 # we create the min and max line
 foreach {y i} [list $VAR($Id,y_axis_min) $VAR($Id,y_data_max) $VAR($Id,y_axis_max) $VAR($Id,y_data_min)] \
 {
  eval .$VAR($Id,Can) create text $VAR($Id,y_axis_text_xpos) $y -text [format "%.6f" $i] -fill black -anchor e -tag y_axis_text
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $VAR($Id,x_axis_min) $y $VAR($Id,x_axis_max) $y -fill black -width 1.000000 -capstyle round -tag y_axis_grid
  }
 }
 
 # if only two lines are wished (min and max) we have already done our job
 if {$VAR($Id,y_axis_line_cnt) == 2} {return}
 
 # the rise of the value of lettering
 set tmp [expr {1.0 * $VAR($Id,y_data_range) / ($VAR($Id,y_axis_line_cnt) - 1)}]
 
 for {set i 1} {$i < ($VAR($Id,y_axis_line_cnt)-1)} {incr i}\
 {
  # the y position for the lettering and the lines
  set y [expr {$i * $VAR($Id,delta_y_axis_lines) + $VAR($Id,y_axis_min)}]
  
  # the value of the lettering
  set value [expr {-$tmp * $i + $VAR($Id,y_data_max)}]
  
  eval .$VAR($Id,Can) create text $VAR($Id,y_axis_text_xpos) $y -text [format "%.6f" $value] -fill black -anchor e -tag y_axis_text
  
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $VAR($Id,x_axis_min) $y $VAR($Id,x_axis_max) $y -fill gray -width 1.000000 -capstyle round -tag y_axis_grid
   .$VAR($Id,Can) create line $VAR($Id,x_axis_min) $y $VAR($Id,y_axis_line_xpos) $y -fill black -width 1.000000 -capstyle round -tag y_axis_line
  }
 }
}


# ---------------------------------------------- 
# draws the linear x-axis and the vertical grid
# ---------------------------------------------- 
proc ::FR_diagram2::draw_x_axis {Id x_min x_max {lettering_only 0}} \
{
 variable VAR
 
 set VAR($Id,x_data_min) $x_min
 set VAR($Id,x_data_max) $x_max
 set VAR($Id,x_data_range) [expr {$x_max - $x_min}]

 # we create the min and max line
 foreach {x i} [list $VAR($Id,x_axis_min) $x_min $VAR($Id,x_axis_max) $x_max] \
 {
  eval .$VAR($Id,Can) create text $x $VAR($Id,x_axis_text_ypos) -text [format "%.6f" $i] -fill black -anchor center -tag x_axis_text
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_min) $x $VAR($Id,y_axis_max) -fill black -width 1.000000 -capstyle round -tag x_axis_grid
  }
 }

 # if only two lines are wished (min and max) we have already done our job 
 if {$VAR($Id,x_axis_line_cnt) == 2} {return}
 
 # the rise of the value of lettering
 set tmp [expr {1.0 * $VAR($Id,x_data_range) / ($VAR($Id,x_axis_line_cnt) - 1)}]
 
 for {set i 1} {$i < ($VAR($Id,x_axis_line_cnt)-1)} {incr i}\
 {
  # the x position for the lettering and the lines
  set x [expr {$i * $VAR($Id,delta_x_axis_lines) + $VAR($Id,x_axis_min)}]
  
  # the value of the lettering
  set value [expr {$tmp * $i + $x_min}]
  
  eval .$VAR($Id,Can) create text $x $VAR($Id,x_axis_text_ypos) -text [format "%.6f" $value] -fill black -anchor center -tag x_axis_text
  
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_min) $x $VAR($Id,y_axis_max) -fill gray -width 1.000000 -capstyle round -tag x_axis_grid
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_max) $x $VAR($Id,x_axis_line_ypos) -fill black -width 1.000000 -capstyle round -tag x_axis_line
  }
 }
}


# ----------------------------------------------------
# draws the logarithmical x-axis and the vertical grid
# ----------------------------------------------------
proc ::FR_diagram2::draw_log_x_axis {Id x_min x_max {lettering_only 0}} \
{
 variable can; variable VAR
 
 # Falk
 # the zoom proc overtakes logarthimical values
# if {$zoom} \
# {
#  set x_min [expr {pow(10,$x_min)}]
#  set x_max [expr {pow(10,$x_max)}]
# }
 
 
 # this looks for the segment in which is value
 # and gives back the start value of this segment
 proc segment {value} \
 {
  if {$value < 10}      {return 1}
  if {$value < 100}     {return 10}
  if {$value < 1000}    {return 100}
  if {$value < 10000}   {return 1000}
  if {$value < 100000}  {return 10000}
 }
 
 
 # we set x_data_min to the next valid element in the segment
 set x_min_seg_start [segment $x_min]
 set x_min_seg_end [expr {$x_min_seg_start * 10}]
 
 set tmp 0
 
 for {set i $x_min_seg_start} {$i <= $x_min_seg_end} {incr i $x_min_seg_start} \
 {
  if {$i == $x_min} {break}
  if {$i > $x_min} {set tmp $x_min_seg_start; break}
 }
 
 # if i > x_min so we are over the limit
 # we go one step back
 set i [expr {$i - $tmp}]
 
 set x_data_min_lin $i
 set VAR($Id,x_data_min) [expr {log10($i)}]
 
 # we set x_data_max to the next valid element in the segment
 set x_max_seg_start [segment $x_max]
 set x_max_seg_end [expr {$x_max_seg_start * 10}]
 
 for {set i $x_max_seg_start} {$i <= $x_max_seg_end} {incr i $x_max_seg_start} \
 {
  if {$i >= $x_max} {break}
 }
 
 if {$i == $x_data_min_lin} {set i [expr{$i + $x_max_seg_start}]}
 
 set x_data_max_lin $i
 set VAR($Id,x_data_max) [expr {log10($i)}]
 
 set VAR($Id,x_data_range) [expr {$VAR($Id,x_data_max) - $VAR($Id,x_data_min)}]
 
 # if value > 1000 we replace this through k(ilo)
 proc transfer_label {value} \
 {
  set tmp [expr {$value / 1000}]
  
  if {$tmp >= 1} {return "${tmp}k"} else {return $value}
 }
 
 # we create the min and max line
 foreach {x i} [list $VAR($Id,x_axis_min) $x_data_min_lin $VAR($Id,x_axis_max) $x_data_max_lin] \
 {
  set label_text [transfer_label $i]
  eval .$VAR($Id,Can) create text $x $VAR($Id,x_axis_text_ypos) -text $label_text -fill black -anchor center -tag x_axis_text
  
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_min) $x $VAR($Id,y_axis_max) -fill black -width 1.000000 -capstyle round -tag x_axis_grid
  }
 }

 set run 1
 set x_axis_values ""
 
 for {set i $x_min_seg_start} {($i <= $x_max_seg_start) && $run} {set i [expr {$i * 10}]} \
 {
  set end [expr {$i * 10}]
  for {set j $i} {$j < $end} {incr j $i} \
  {
   if {$j <= $x_data_min_lin} {continue}
   if {$j >= $x_data_max_lin} {set run 0; break}
   
   lappend x_axis_values $j
  }
 }

 set delta_x_axis [expr {$VAR($Id,x_axis_range) / ($VAR($Id,x_data_max) - $VAR($Id,x_data_min))}]

 foreach i $x_axis_values \
 {
  set x [expr {$delta_x_axis * (log10($i) - $VAR($Id,x_data_min)) + $VAR($Id,x_axis_min)}]
  
  set run 1
  
  # we don't write the 7 and 9 label (to save space on the axis) 
  if {![expr {$i % 7}] || ![expr {$i % 9}]} {set run 0}
  
  if {$run} \
  {
   eval .$VAR($Id,Can) create text $x $VAR($Id,x_axis_text_ypos) -text [transfer_label $i] -fill black -anchor center -tag x_axis_text
  }
  
  if {!$lettering_only} \
  {
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_min) $x $VAR($Id,y_axis_max) -fill gray -width 1.000000 -capstyle round -tag x_axis_grid
   .$VAR($Id,Can) create line $x $VAR($Id,y_axis_max) $x $VAR($Id,x_axis_line_ypos) -fill black -width 1.000000 -capstyle round -tag x_axis_line
  }
 }
}


# ---------------------------------------------- 
# this plots the data
# ---------------------------------------------- 
proc ::FR_diagram2::draw_curve {Id data_range} \
{
 variable VAR; variable DATA
 
 # if no data range is specified we havn't anything to do...
 if {[llength $data_range] !=  2} {return 0}
 
 set VAR($Id,data_range) $data_range
 
 set delta_x_data [expr {$VAR($Id,x_axis_range) / $VAR($Id,x_data_range)}]
 set delta_y_data [expr {$VAR($Id,y_axis_range) / $VAR($Id,y_data_range)}]
 
 # with i we make a delay of one values pairs to have start and end value
 set i 0
 
 foreach {x y} [lrange $DATA($Id) [lindex $VAR($Id,data_range) 0] [lindex $VAR($Id,data_range) 1]] \
 {
  if {$VAR($Id,x_axis_log)} \
  {
   set x [expr {log10($x)}]
  }
  
  set x_end $x
  set y_end $y
  
  # in the first loop we don't come in this body, so we can define x_start
  # and y_start at the end of this foreach body
  if {$i} \
  {
   #we transfere the xy-data to xy-canvas
   set x_start_can [expr {$delta_x_data * ($x_start - $VAR($Id,x_data_min)) + $VAR($Id,x_axis_min)}]
   set x_end_can [expr {$delta_x_data * ($x_end - $VAR($Id,x_data_min)) + $VAR($Id,x_axis_min)}]
   
   set y_start_can [expr {$delta_y_data * ($VAR($Id,y_data_max) - $y_start) + $VAR($Id,y_axis_min)}]
   set y_end_can [expr {$delta_y_data * ($VAR($Id,y_data_max) - $y_end) + $VAR($Id,y_axis_min)}]
   
   #we check if the xy-canvas are outside of the diagram
   set y_start_outside 0; set y_end_outside 0
   if {($y_start_can > $VAR($Id,y_axis_max)) || ($y_start_can < $VAR($Id,y_axis_min))} {set y_start_outside 1}
   if {($y_end_can > $VAR($Id,y_axis_max)) || ($y_end_can < $VAR($Id,y_axis_min))} {set y_end_outside 1}
   
   set draw_line 0
   
   # if the start_values is inside and the end_value is outside
   if {(!$y_start_outside && $y_end_outside)} \
   {
    # we save the y_end_can value for the calculation of x_end_can
    set y_end_can_real $y_end_can
    
    if {$y_end_can > $VAR($Id,y_axis_max)} \
    {
     set y_end_can $VAR($Id,y_axis_max)
    } else \
    {
     # if y_end_can < $y_axis_min because we checked before for to be outside
     set y_end_can $VAR($Id,y_axis_min)
    }
    
    # we set the x_end values to the place where the line goes out (y_axis_max or y_axis_min)
    set x_end_can [expr {($y_end_can - $y_start_can) * ($x_end_can - $x_start_can) / ($y_end_can_real - $y_start_can) + $x_start_can}]
    
    
    set draw_line 1
   }
   
   # if the start_values is outside and the end_value is inside
   if {($y_start_outside && !$y_end_outside)} \
   {
    # we save the y_start_can value for the calculation of x_start_can
    set y_start_can_real $y_start_can
    
    if {$y_start_can > $VAR($Id,y_axis_max)} \
    {
     set y_start_can $VAR($Id,y_axis_max)
    } else \
    {
     #y_start_can must be < $y_axis_min because we checked before for to be outside
     set y_start_can $VAR($Id,y_axis_min)
    }
    
    # we set the x_start values to place where the lines comes in (y_axis_max or y_axis_min)
    set x_start_can [expr {($y_start_can - $y_start_can_real) * ($x_end_can - $x_start_can) / ($y_end_can - $y_start_can_real) + $x_start_can}]
    
    set draw_line 1
   }
       
   #if both values are inside in the diagram (the normal case)
   if {(!$y_start_outside && !$y_end_outside) || $draw_line} \
   {
    .$VAR($Id,Can) create line $x_start_can $y_start_can $x_end_can $y_end_can -fill blue -width 1.000000 -capstyle round -tag data_curve
   }
  }
  
  set x_start $x_end
  set y_start $y_end
  set i 1 
 }
}


# ---------------------------------------------- 
# this add a title to the diagram
# ---------------------------------------------- 
proc ::FR_diagram2::diagram_title {Id {title ""}} \
{
 variable VAR
 
 ::FR_diagram2::delete $Id diagram_title
 
 if {$title ne ""} {set VAR($Id,diagram_title) $title}
 
 set diagram_title [string range $VAR($Id,diagram_title) 0 99]
 
 .$VAR($Id,Can) create text $VAR($Id,diagram_title_xpos) $VAR($Id,diagram_title_ypos) -text $diagram_title -fill black -anchor center -tag diagram_title
}


# ---------------------------------------------- 
# delete tag-list in can
# ---------------------------------------------- 
proc ::FR_diagram2::delete {Id {tag ""}} \
{
 variable VAR
 
 if {$tag eq ""} {return}

 foreach i $tag {.$VAR($Id,Can) delete $i}
}


# ---------------------------------------------- 
# we check if x_max > x_min and y_max > y_min
# ---------------------------------------------- 
proc ::FR_diagram2::check_limits {x_min x_max y_min y_max} \
{ 
 set limits_error 0
 
 if {$x_max <= $x_min} \
 {
  set limits_error 1
  return $limits_error
 }
 
 if {$y_max < $y_min} \
 {
  set limits_error 1
  return $limits_error
 }
 
 if {$y_max == $y_min} \
 {
  set limits_error 2
  return $limits_error
 }
 
 return $limits_error
}


# ---------------------------------------------- 
# saves the start of the rectangle zoom box
# ---------------------------------------------- 
proc ::FR_diagram2::start_zoom_box {Id x y} \
{
 variable VAR 
 
 set VAR($Id,start_zoom_box) 1
 set VAR($Id,x_start_zoom_box) $x
 set VAR($Id,y_start_zoom_box) $y
}


# ---------------------------------------------- 
# saves the end of the rectangle zoom box
# ---------------------------------------------- 
proc ::FR_diagram2::end_zoom_box {Id x y} \
{
 variable VAR
 
 set VAR($Id,start_zoom_box) 0
 set VAR($Id,x_end_zoom_box) $x
 set VAR($Id,y_end_zoom_box) $y
 
 ::FR_diagram2::delete $Id zoom_box 

 ::FR_diagram2::check_zoom_box $Id
 ::FR_diagram2::zoom_calc $Id
}


# ---------------------------------------------- 
# draws the rectangle zoom box
# ---------------------------------------------- 
proc ::FR_diagram2::draw_zoom_box {Id x y} \
{
 variable VAR
 
 if {![info exists VAR($Id,start_zoom_box)] || !$VAR($Id,start_zoom_box)} {return}
 
 ::FR_diagram2::delete $Id zoom_box 
 
 .$VAR($Id,Can) create rectangle $VAR($Id,x_start_zoom_box) $VAR($Id,y_start_zoom_box) $x $y -outline black -tags zoom_box 
}


# ---------------------------------------------- 
# checks and corrects if the zoom box is outside of the diagram
# we also check if the box drawn from right to left 
# ---------------------------------------------- 
proc ::FR_diagram2::check_zoom_box {Id} \
{
 variable VAR
 
 # first we limit zoom box coordinates to the min/max axis values 
 if {$VAR($Id,x_start_zoom_box) < $VAR($Id,x_axis_min)} {set VAR($Id,x_start_zoom_box) $VAR($Id,x_axis_min)}
 if {$VAR($Id,x_end_zoom_box) > $VAR($Id,x_axis_max)} {set VAR($Id,x_end_zoom_box) $VAR($Id,x_axis_max)}
 
 if {$VAR($Id,y_start_zoom_box) < $VAR($Id,y_axis_min)} {set VAR($Id,y_start_zoom_box) $VAR($Id,y_axis_min)}
 if {$VAR($Id,y_end_zoom_box) > $VAR($Id,y_axis_max)} {set VAR($Id,y_end_zoom_box) $VAR($Id,y_axis_max)}
 
 # we check if the box drawn from right to left
 if {$VAR($Id,x_start_zoom_box) > $VAR($Id,x_end_zoom_box)} \
 {
  set tmp $VAR($Id,x_start_zoom_box)
  set VAR($Id,x_start_zoom_box) $VAR($Id,x_end_zoom_box)
  set VAR($Id,x_end_zoom_box) $tmp
  
  set tmp $VAR($Id,y_start_zoom_box)
  set VAR($Id,y_start_zoom_box) $VAR($Id,y_end_zoom_box)
  set VAR($Id,y_end_zoom_box) $tmp
  
  unset tmp
 }
}


# ---------------------------------------------- 
# we calculate back from xy-canvas to xy-data
# ---------------------------------------------- 
proc ::FR_diagram2::zoom_calc {Id} \
{
 variable VAR; variable DATA

 set tmp [expr {$VAR($Id,x_data_range) / $VAR($Id,x_axis_range)}]
 
 set x_min [expr {($VAR($Id,x_start_zoom_box) - $VAR($Id,x_axis_min)) * $tmp + $VAR($Id,x_data_min)}]
 set x_max [expr {($VAR($Id,x_end_zoom_box) - $VAR($Id,x_axis_min)) * $tmp + $VAR($Id,x_data_min)}]
 
 set tmp [expr {$VAR($Id,y_data_range) / $VAR($Id,y_axis_range)}] 
 
 set y_max [expr {$VAR($Id,y_data_max)-($VAR($Id,y_start_zoom_box) - $VAR($Id,y_axis_min)) * $tmp}]
 set y_min [expr {$VAR($Id,y_data_max)-($VAR($Id,y_end_zoom_box) - $VAR($Id,y_axis_min)) * $tmp}]
 
 # if x_max <= x_min or y_max <= y_min we can't further zoom
 set limits_error [::FR_diagram2::check_limits $x_min $x_max $y_min $y_max]
 if {$limits_error} {return}
 
 # we select out the needed x-values (from data) and start the drawing
 if {[llength $VAR($Id,data_range)] !=  2} {return 0}
 
 set start_index [lindex $VAR($Id,data_range) 0]
 set end_index [lindex $VAR($Id,data_range) 1]
 
 #we are looking for x > minimum
 for {set i $start_index} {$i < $end_index} {incr i 2} \
 {
  set tmp [lindex $DATA($Id) $i]
  
  if {$VAR($Id,x_axis_log)} {set tmp [expr {log10($tmp)}]}
  
  if {$tmp >= $x_min} {break}
 }
 
 #we are looking for x < maximum
 for {set j $i} {$j < $end_index} {incr j 2} \
 {
  set tmp [lindex $DATA($Id) $j]
  
  if {$VAR($Id,x_axis_log)} {set tmp [expr {log10($tmp)}]}
  
  if {$tmp > $x_max} {break}
 }
 
 # with the break command we are one element-pair to far
 # only -1 because we need also the y-value at x_data_max
 incr j -1
 
 # we need at least three x (and three y) values for further zooming
 if {[expr {abs($i - $j)}] < 5} {return}
 
 # we save the indeces i and j and append it to the zoom history
 set VAR($Id,data_range) [list $i $j]
 lappend VAR($Id,zoom_history) $i $j
 
 ::FR_diagram2::zoom_draw $Id $y_min $y_max
}


# ---------------------------------------------- 
# draw the chosen zoom
# ---------------------------------------------- 
proc ::FR_diagram2::zoom_draw {Id y_min y_max} \
{
 variable VAR; variable DATA
 
 #x_min is the first and x_max the last point of data
 set start_index [lindex $VAR($Id,data_range) 0]
 set end_index [lindex $VAR($Id,data_range) 1]
 
 set x_min [lindex $DATA($Id) $start_index]
 set x_max [lindex $DATA($Id) [expr {$end_index - 1}]]
 
 # now we draw all
 if {$VAR($Id,x_axis_log)} \
 {
  set tmp [list data_curve x_axis_text x_axis_grid x_axis_line y_axis_text]
 } else \
 {
  set tmp [list data_curve x_axis_text y_axis_text]
 }
 
 ::FR_diagram2::delete $Id $tmp
 
 ::FR_diagram2::draw_y_axis $Id $y_min $y_max 1
 
 if {$VAR($Id,x_axis_log)} \
 {
  ::FR_diagram2::draw_log_x_axis $Id $x_min $x_max 1
 } else \
 {
  ::FR_diagram2::draw_x_axis $Id $x_min $x_max 1
 }
 
 ::FR_diagram2::draw_curve $Id $VAR($Id,data_range)
}


# ---------------------------------------------- 
# the zoom out function
# ---------------------------------------------- 
proc ::FR_diagram2::zoom_out {Id view_all} \
{
 variable DATA; variable VAR
 
 if {$view_all == 0} {
  # this is the normal zoom out function
  
  # we check if the zoom_history has already reached the full picture
  if {[llength $VAR($Id,zoom_history)] < 3} {return}
 
  # if the list length of data smaller as the wished index (maximum)
  # we abort the function, (end-2) because we want the last plot (and not the current plot)
  if {[llength $DATA($Id)] < [lindex $VAR($Id,zoom_history) end-2]} {return}
 
  set VAR($Id,data_range) [list [lindex $VAR($Id,zoom_history) end-3] [lindex $VAR($Id,zoom_history) end-2]]
 } else {
  # this is the view all function 
  set VAR($Id,data_range) [list 0 $VAR($Id,last_element)]
 }
 
 # we are searching the highest and the lowest y value
 set start_index [lindex $VAR($Id,data_range) 0]
 # we are interested in the y-values
 incr start_index
 set end_index [lindex $VAR($Id,data_range) 1]
 
 set y_min [lindex $DATA($Id) $start_index]
 set y_max $y_min
 
 for {set i $start_index} {$i <= $end_index} {incr i 2} \
 {
  set tmp [lindex $DATA($Id) $i]
  
  if {$tmp > $y_max} \
  {
   set y_max $tmp
   continue
  }
   
  if {$tmp < $y_min} \
  {
   set y_min $tmp
  }
 } 
   
 ::FR_diagram2::zoom_draw $Id $y_min $y_max
 
 if {$view_all == 0} {
  # now we drew one zoom out so can delete the last zoom
  set VAR($Id,zoom_history) [lrange $VAR($Id,zoom_history) 0 end-2]
 } else {
  # we drew all datas (view_all==1) so we can delete the whole zoom history
  set VAR($Id,zoom_history) [list 0 $VAR($Id,last_element)]
 }
}


# ---------------------------------------------- 
# we will wait until the last configure event
# ---------------------------------------------- 
proc ::FR_diagram2::redraw_request {Id} \
{
 variable VAR
 
 if {$VAR($Id,redraw_request) eq ""} \
 {
  set VAR($Id,redraw_request) [after idle [list ::FR_diagram2::redraw $Id]]
 }
}


# ------------------------------------------------------------------
# the redraw function in the case of a resizing of the canvas widget
# ------------------------------------------------------------------
proc ::FR_diagram2::redraw {Id} \
{
 variable VAR

 set VAR($Id,canvas_x_size_old) $VAR($Id,canvas_x_size)
 set VAR($Id,canvas_y_size_old) $VAR($Id,canvas_y_size)
 
 update
 
 ::FR_diagram2::get_canvas_size $Id
 
 if {($VAR($Id,canvas_x_size_old) != $VAR($Id,canvas_x_size)) || ($VAR($Id,canvas_y_size_old) != $VAR($Id,canvas_y_size))} \
 {
  ::FR_diagram2::redraw_sub $Id
 }
 
 set VAR($Id,redraw_request) ""
}

proc ::FR_diagram2::redraw_sub {Id} \
{
 variable VAR
 
 ::FR_diagram2::init $Id
 ::FR_diagram2::look $Id
 ::FR_diagram2::draw_y_axis $Id $VAR($Id,y_data_min) $VAR($Id,y_data_max)
 
 if {$VAR($Id,x_axis_log)} \
 {
  ::FR_diagram2::draw_log_x_axis $Id $VAR($Id,x_data_min) $VAR($Id,x_data_max)
 } else \
 {
  ::FR_diagram2::draw_x_axis $Id $VAR($Id,x_data_min) $VAR($Id,x_data_max)
 }
 
 if {[llength $VAR($Id,data_range)] ==  2} \
 {
  ::FR_diagram2::draw_curve $Id $VAR($Id,data_range)
 }
 
 if {$VAR($Id,diagram_title) ne ""} \
 {
  ::FR_diagram2::diagram_title $Id
 }
}


# ---------------------------------------------- 
# we analyse and draw the data
# ---------------------------------------------- 
proc ::FR_diagram2::draw_data {Id Data {X_Axis_Log 0}} \
{
 variable DATA; variable VAR
 
 # we check: is the data list valid
 set data_length [llength $Data]
 if {$data_length < 4} {return 0}
 
 set DATA($Id) $Data
 set VAR($Id,last_element) [expr {$data_length - 1}]
 set VAR($Id,data_range) [list 0 $VAR($Id,last_element)]
 
 # we set x_min and x_max to the first and last x element in the data list
 set x_min [lindex $Data 0]
 set x_max [lindex $Data end-1]
 
 # we are searching the highest and the lowest y value
 set y_min [lindex $Data 1]
 set y_max $y_min
 
 for {set i 1} {$i < $data_length} {incr i 2} \
 {
  set tmp [lindex $Data $i]
  
  if {$tmp > $y_max} \
  {
   set y_max $tmp
   continue
  }
   
  if {$tmp < $y_min} \
  {
   set y_min $tmp
  }
 }
 
 # if x_max <= x_min or y_max < y_min we can't draw the curve
 set limits_error [::FR_diagram2::check_limits $x_min $x_max $y_min $y_max]
 if {$limits_error == 1} {return}
 if {$limits_error == 2} \
 {
  set y_min [expr {$y_min - 1}]
  set y_max [expr {$y_max + 1}]
 }
 
 # now we can make the values to official values
 set VAR($Id,x_data_min) $x_min
 set VAR($Id,x_data_max) $x_max
 set VAR($Id,y_data_min) $y_min
 set VAR($Id,y_data_max) $y_max 
 
 # we reset the history of the last zooms
 set VAR($Id,zoom_history) [list 0 $VAR($Id,last_element)]

 
 # we delete the old things in the diagram
 if {$X_Axis_Log != $VAR($Id,x_axis_log)} \
 { 
  set tmp [list data_curve x_axis_text x_axis_grid x_axis_line y_axis_text]
 } else \
 {
  set tmp [list data_curve x_axis_text y_axis_text]
 }
 
 ::FR_diagram2::delete $Id $tmp
 
 # we draw the y axis new 
 ::FR_diagram2::draw_y_axis $Id $VAR($Id,y_data_min) $VAR($Id,y_data_max) 1
 
 # we draw the x axis new
 if {$X_Axis_Log} \
 {
  set VAR($Id,x_axis_log) 1
  ::FR_diagram2::draw_log_x_axis $Id $VAR($Id,x_data_min) $VAR($Id,x_data_max) 0
 } else \
 {
  set VAR($Id,x_axis_log) 0
  ::FR_diagram2::draw_x_axis $Id $VAR($Id,x_data_min) $VAR($Id,x_data_max) 0
 }
 
 # we draw the data points new
 ::FR_diagram2::draw_curve $Id $VAR($Id,data_range)
}


# ---------------------------------------------- 
# calls only the draw function with the log flag
# ----------------------------------------------
proc ::FR_diagram2::draw_log_data {Id Data} \
{ 
  ::FR_diagram2::draw_data $Id $Data 1
}


#package require Tk
#toplevel .
#frame .diagram
#pack .diagram

#set data [list 0 0.10827637 0.1 0.12744141 0.2 0.12310791 0.3 0.095184326 0.4 0.050842285 0.5 -0.0033874512 0.6 -0.057525635 0.7 -0.10153198 0.8 -0.12789917 0.9 -0.13189697 1 -0.11227417]
#set data [list 500 0.1 1000 0.1 2000 0.5 2500 0.5 3000 1.0 8000 -1.00]

#set DiaId [::FR_diagram2::create -path . -x_axis_min 500 -x_axis_max 8000 -x_axis_log 1 -data $data]
#set DiaId [::FR_diagram2::create -path .diagram -data $data]
#::FR_diagram2::diagram_title $DiaId "Falks Diagramm"

