#*****************************************************************************
# Script process a table of content read out from an audio CD with cdparanoia. 
# It returns a list with all titles and generates the disc-id for the 
# GnuDB-database access 
#
# The script bases on the package discid 0.1 of Joachim Heidemeier from 2011
# and got some information from the C-program cd-discid written by 
# Robert Woodcock
#
# Author: Falk Richter, falk.richter@yandex.com, 2021
#
# $Id: gnudb.tcl,v 1.3 2021/07/07 21:33:58 falk Exp falk $
#***************************************************************************** 

namespace eval ::cddb {
  variable Local
  # we need only this functions from the outside
  namespace export ProcessToc GetErrorMsg 
}


########################################################### 
# Converts the string mm:ss.ff into a list with three 
# elements (minutes seconds milliseconds),
# 2 sec lead-in time (or lead out time in case of the total
# CD length) are automatically added
###########################################################
proc ::cddb::ConvertTime { TimeString } {
  
  # split the mm:ss.ff format into a list with three elements
  set TimeStringList [split $TimeString {: .}]
	
  # get the elements again
  set Minutes [lindex $TimeStringList 0]
  set Seconds [lindex $TimeStringList 1]
  set MSeconds [lindex $TimeStringList 2] 
  
  # remove the leading zeros and convert all to integer
  scan $Minutes %d Minutes
  scan $Seconds %d Seconds
  scan $MSeconds %d MSeconds
	
  # add 2s lead in/out time
  incr Seconds 2
	
  # check for seconds "overflow"
  if {$Seconds > 60} {
    # we increase the minutes and save the rest in the seconds
    # field, TimeList is our return value
    set TimeList [list [incr Minutes] [expr {$Seconds - 60}] $MSeconds]
  } else {
    # no overflow in seconds, we can simple save all three values
	set TimeList [list $Minutes $Seconds $MSeconds]
  }
  
  return $TimeList
}


###########################################################
# Parse the table of content (toc) of the audio CD
###########################################################
proc ::cddb::ParseToc {} {
  variable Local;
  
  # an example for a TOC generated by cdparanoia:
  #
  # Table of contents (audio tracks only):
  # track        length               begin        copy pre ch
  # ===========================================================
  #   1.    16799 [03:43.74]        0 [00:00.00]    no   no  2
  #   2.    13364 [02:58.14]    16799 [03:43.74]    no   no  2
  #  ...
  #  13.    22035 [04:53.60]   205230 [45:36.30]    no   no  2
  # TOTAL  227265 [50:30.15]    (audio only)

  # skip the header from cdparanoia, 
  # header ends with a line of equal signs (==...) 
  
  # index for already parsed lines
  set Idx 0
  set Lines [split $Local(RawToc) \n]
  set LineCnt [llength $Lines] 

  if {$LineCnt == 0} {
    set Local(ErrorFlag) 1
    set Local(ErrorMsg) [msgcat::mc "The table of content is empty."]
    set Local(ErrorDetail) "$Local(RawToc)"	
	return $Local(ErrorFlag)
  }
  
  # first we eat the header
  foreach Line $Lines {
    # the index is used later to skip the header
    incr Idx;
    # we are looking for the line of equal signs 
    if {[regexp {[=]+} $Line]} {
      # we have found the start of the TOC
      break;
    }
  }

  # if we are at the end of the list we didn't find
  # the start of the TOC
  if { $Idx == $LineCnt } {
    set Local(ErrorFlag) 1
    set Local(ErrorMsg) [msgcat::mc "The table of content isn't valid."]
    set Local(ErrorDetail) "$Local(RawToc)"	
	return $Local(ErrorFlag)
  }
	  
  # we start without expectations
  set Local(TrackCnt) 0
  set Local(TrackList) ""
  
  # try to get the tracks behind the header
  for {} {$Idx < $LineCnt} {incr Idx} {
	set Line [lindex $Lines $Idx]
	if {[regexp -all -- \
 	{([0-9]+)\..*([0-9]{2}:[0-9]{2}\.[0-9]{2}).*([0-9]{2}:[0-9]{2}\.[0-9]{2}).*}\
	$Line All Track Length Begin]} {
	  # add the track to the list of tracks (needed for the rip task)
	  lappend Local(TrackList) $Track
	  # the last track sets the count
	  set Local(TrackCnt) $Track
      # get tracks, indexed from 0 
	  set Local([expr {$Track - 1}]) [ConvertTime $Begin]
	} elseif {[regexp -all -- \
	  {TOTAL.*([0-9]{2}:[0-9]{2}\.[0-9]{2})} $Line All Length]} {
	  # get total_tracks time
      set Local(TotalLength)  [ConvertTime $Length]
      set Local(TotalSeconds) \
	  [expr {[lindex $Local(TotalLength) 0] * 60 + [lindex $Local(TotalLength) 1]}]
	  # we have all information
	  return 0
    } else {
      set Local(ErrorFlag) 1
      set Local(ErrorMsg) [msgcat::mc "Found an unexpected line in the table of content."]
      set Local(ErrorDetail) "$Line"	
	  return $Local(ErrorFlag)
	}
  }
}


###########################################################
# Calculates the sum over all digits of a number
###########################################################
proc ::cddb::DigitsSum {Number} {
  set Result 0
  
  while {$Number > 0} {
	# we add the (lowest) digit on the right side
    set Result [expr {$Result + ($Number % 10)}]
	# remove the lowest digit from the number
	# (integer cast to avoid a float calculation)
	set Number [expr {int($Number / 10)}]
  } 
  return $Result 
}


###########################################################
# Calculates the disc-id
###########################################################
proc ::cddb::CalcID {} {
  variable Local
  set n 0
   
  # we calculate the total of the digits of the track length
  # in seconds, we do it as summary over all tracks
  for {set i 0} {$i < $Local(TrackCnt)} {incr i} {
	 set Minutes [lindex $Local($i) 0]
	 set Seconds [lindex $Local($i) 1]
	 set TotalSeconds [expr {($Minutes * 60) + $Seconds}]
	 set n [expr {$n + [DigitsSum $TotalSeconds]}]
  }
  
  # FirstSongTotalSecs should be in the most cases 2s from the lead-in 
  # time (the start of the first song is 00:00.00 on a pure audio CD)
  set FirstSongTotalSecs [expr {[lindex $Local(0) 0] * 60 + [lindex $Local(0) 1]}]
  
  set TimeDiff [expr {$Local(TotalSeconds) - $FirstSongTotalSecs}]
  set Id [expr {($n % 0xff) << 24 | $TimeDiff << 8 | $Local(TrackCnt)}]
  set Local(ID) [format %08x $Id]
}


###########################################################
# Calculates the frame count from the time information
# (each second has 75 frames)
###########################################################
proc ::cddb::FrameOffsets {} {
  variable Local

  # reset the list
  set Local(FrameOffsets) ""
  
  for {set i 0} {$i < $Local(TrackCnt)} {incr i} {
	# fetch the offset from each track
    set TimeOffset $Local($i)
	# 75 frames per second, calculate the frame offset from the minutes,
	# seconds and mseconds
	lappend Local(FrameOffsets) [expr {([lindex $TimeOffset 0] * 60 * 75) + \
	([lindex $TimeOffset  1] * 75) + [lindex $TimeOffset 2]}] 
  }
}


###########################################################
# Parse the table of content of an audio CD and calculate
# the disc-id
###########################################################
proc ::cddb::ProcessToc { RawToc } {
  variable Local
  
  # reset the module global error flag
  set Local(ErrorFlag) 0
  
  # copy the raw toc into the module global variable
  set Local(RawToc) $RawToc

  # check if we could parse the TOC
  if { [ParseToc] == 0 } {
    # the first part for the GnuDB access is the disc id
	CalcID
  
	# the frame offsets are the second part
	FrameOffsets 

	# now we combine all together to one list
	set CddbList [concat $Local(ID) $Local(TrackCnt) $Local(FrameOffsets) $Local(TotalSeconds)]
    
	# the return value contains: the success flag, the track list and the 
	# list for cddb access
	return [list $Local(ErrorFlag) $Local(TrackList) $CddbList]
  } else {
	# something went wrong, we return only the error flag
	return $Local(ErrorFlag)
  }
}


########################################################### 
# Returns the error message and the error details
########################################################### 
proc ::cddb::GetErrorMsg {} {
  variable Local

 return [list $Local(ErrorMsg) $Local(ErrorDetail)] 
} 

